/********************************************************
 eyeCrop: Version 1.0: Dec., 1999 

 Rescale original face images to have inter-eye distance = eyeDist0 pixels. 
 Eye distance obtained from ppm training images with a faceParams comment
 line.  This comment line the provides the pixel coords of the
 eye centers and the nose tip.  These were obtained by hand.

 Linear interp used for rescaling (SHOULD USE FILTERING AND DOWN-SAMPLING). 

 To compile:
  make eyeCrop
********************************************************/

# include "macros.h"
# include "utils.h"
# include "endianness.h"
# include "imageFile-io.h"
# include "mapping.h"
# include "faceImageIO.h"

unsigned char grey = 127;
int eyeRadX = 10;
int eyeUpY = 15;
int eyeDownY = 10;

void cropWritePfm(char *path, int *cropBox, float *imageFlt,
		  int nx, int ny)
{
  float *cropImage;
  unsigned char *byteImage;
  int nxC, nyC, ix, iy, ixy, ixc, iyc;

  /* Allocate and initialize cropped image */
  nxC = cropBox[2] - cropBox[0];
  nyC = cropBox[3] - cropBox[1];
  if (nxC <= 0 || nyC <= 0)
    error("Invalid (empty) crop box.","");
  grabFloatMemory(&cropImage, nxC * nyC, "cropImage");
  for(ixy=0; ixy<nxC * nyC; ixy++)
    cropImage[ixy] = (float) grey;

  /* Set cropped image to the appropriate grey-levels, where available */
  for(iyc=0; iyc<nyC; iyc++) {
    iy = cropBox[1] + iyc;
    if (iy >=0 && iy < ny)
      for(ixc=0; ixc<nxC; ixc++) {
	ix = cropBox[0] + ixc;
	if (ix >=0 && ix < nx)
	  cropImage[index(ixc,iyc, nxC)] = imageFlt[index(ix,iy,nx)];
      }
  }
  save_pfm_image(path, cropImage, nxC, nyC);

  /* DEBUG: Save a pgm image for viewing */
  quantize_float_image(&byteImage, cropImage, nxC, nyC);
  save_pgm_image(path, byteImage, nxC, nyC);
  free(byteImage);
  free(cropImage);
    
}

main(argc, argv)
     int argc;
     char **argv;
{
 
  float *imageFlt, *warped, *cropImage;
  float x_orig[2], warpParam[8], invWarp[8], centPt[2];
  float eyeDist, scl, theta, x_old[2];
  float facePos[6];
  float eyeDist0;
 
  int cropBox[4];
  int nxImage, nyImage;
  int nxWarp, nyWarp;
  int nxCrop, nyCrop;
  int i,k, ix, iy;
  int nTrain, kTrain;
  unsigned char *image, *cimage[3];

  char fnInput[MAXLEN], fnTrain[MAXLEN][MAXLEN],scrName[MAXLEN];
  char rootName[MAXLEN];
  char fnOutput[MAXLEN], pathNameOutput[MAXLEN], outputFileRoot[MAXLEN];
  char pathNameInput[MAXLEN];
  char comment[MAXLEN];
  char inImageType[3];

  FILE *fpIn;

  /******************************/
  /* Read input data from stdin */
  /******************************/
  fprintf(stderr, "Inter-eye distance : ");
  scanf("%e", &eyeDist0);

  fprintf(stderr, "Output file path and rootname: \n");
  if (read_file_path_and_root(pathNameOutput, outputFileRoot) == EOF)
    exit(0);

  /*****************************/
  /* Read training image names */
  /*****************************/
  fprintf(stderr, "Training image file path: \n");
  scanf("%s", pathNameInput);
  fprintf(stderr, 
          "Input a list of training image names (terminate with \"null\"): ");
  nTrain =0;
  while(scanf("%s", fnTrain[nTrain]) != EOF) {
    if (checkForNull(fnTrain[nTrain]) == 0) 
      break;
    nTrain++;
    if (nTrain >= MAXLEN) {
      fprintf(stderr, 
       "Warning: Truncating the training sequence to %d training images\n",
	      nTrain);
      break;
    }
  }
  
  fpIn = NULL;

  for(kTrain=0; kTrain<nTrain; kTrain++) {
    /*******************/
    /* Read image data */
    /*******************/
    strcpy(fnInput, pathNameInput);
    strcat(fnInput, fnTrain[kTrain]);

    /* Open image file */
    if ((fpIn = freopenInputFile(fnInput, fpIn)) == NULL) {
      fprintf(stderr, "Can't open image file %s\n",fnInput);
      exit(0);
    }
    if (stripPnmHeader(fpIn, &nxImage, &nyImage, inImageType) < 0) {
      fprintf(stderr," Image file: %s, type: %s\n", fnInput, inImageType);
      error("Unrecognized image file type",""); 
    }
    rewind(fpIn);
    /* Read face coords and file type */
    if (0> readFaceHeader(fpIn, facePos))
      fprintf(stderr, " cannot read face coords for %s\n", fnInput);
    rewind(fpIn);

    if (inImageType[0]=='P') {
      if (inImageType[1] =='6') {
	read_ppm_image(cimage, fpIn, fnInput, &nxImage, &nyImage);
	image = cimage[0];
	free(cimage[1]);
	free(cimage[2]);
      } else if (inImageType[1] =='5') {
	read_pgm_image(&image, fpIn, fnInput, &nxImage, &nyImage);
      } else
	error("Unrecognized input file type ", inImageType);

      grabFloatMemory(&imageFlt, nxImage*nyImage, "imageFlt");
      for(i=0; i<nxImage*nyImage; i++)
	imageFlt[i] = (float) image[i];
      free(image);
    } else if (inImageType[0]=='F') {
      if (inImageType[1] =='5') {
	read_pfm_image(&imageFlt, fpIn, fnInput, &nxImage, &nyImage);
      } else
	error("Unrecognized input file type ", inImageType);
    } else
      error("Unrecognized input file type ", inImageType);


    /* Compute inter eye distance from image file header comment */
    eyeDist = (facePos[0] - facePos[2]) * (facePos[0] - facePos[2]);
    eyeDist += (facePos[1] - facePos[3]) * (facePos[1] - facePos[3]);
    eyeDist = sqrt(eyeDist);

    /* Rescale face image to have inter-eye distance eyeDist0 */
    for(i=0; i<8;i++)
      warpParam[i] = 0.0;
    theta = atan2(facePos[3]-facePos[1], facePos[2] - facePos[0]);
    scl = eyeDist0/eyeDist;
    warpParam[2] = scl * cos(theta) ;
    warpParam[3] = -scl * sin(theta) ;
    warpParam[4] = -warpParam[3];
    warpParam[5] = warpParam[2];
    x_orig[0] = x_orig[1] = 0.0;
    
    warpAndBound(imageFlt, nxImage, nyImage, 
		 &warped, &nxWarp, &nyWarp,
		 x_orig, warpParam, 128, 1, invWarp);
    fprintf(stderr, "scl = %f, input image size %d %d, output size %d %d\n",
	    scl, nxImage, nyImage, nxWarp, nyWarp);
    free(imageFlt);
    imageFlt = warped;
    nxImage = nxWarp;
    nyImage = nyWarp;

    /* Get warped face positions */
    if (-1 == invertRational(invWarp, warpParam))
      error("Rational transform for prewarp nearly singular","");
    for(k=0; k<3; k++) {
      x_old[0] = facePos[2*k];
      x_old[1] = facePos[2*k+1];
      rationalPtSample(x_old, x_orig, warpParam, centPt); 
      facePos[2*k] = centPt[0];
      facePos[2*k+1] = centPt[1];
    }

    /* Find cropBox for eyes from warped face positions */
    scl = 1.0;
    cropBox[0] = ROUND(facePos[0]-eyeRadX*scl); 
    cropBox[1] = ROUND(facePos[1]-eyeUpY*scl);
    cropBox[2] = cropBox[0] + ROUND(2.0 *eyeRadX*scl);
    cropBox[3] = cropBox[1] + ROUND((eyeUpY+eyeDownY)*scl);
    sprintf(scrName, "%s%s_reye", pathNameOutput, outputFileRoot);
    build_filename(fnOutput, scrName, 3, kTrain, "");
    cropWritePfm(fnOutput, cropBox, warped, nxWarp, nyWarp);

    cropBox[0] = ROUND(facePos[2]-eyeRadX*scl); 
    cropBox[1] = ROUND(facePos[3]-eyeUpY*scl);
    cropBox[2] = cropBox[0] + ROUND(2.0 *eyeRadX*scl);
    cropBox[3] = cropBox[1] + ROUND((eyeUpY+eyeDownY)*scl);
    sprintf(scrName, "%s%s_leye", pathNameOutput, outputFileRoot);
    build_filename(fnOutput, scrName, 3, kTrain, "");
    cropWritePfm(fnOutput, cropBox, warped, nxWarp, nyWarp);

    cropBox[0] = ROUND(facePos[4]-eyeRadX*scl); 
    cropBox[1] = ROUND(facePos[5]-eyeUpY*scl);
    cropBox[2] = cropBox[0] + ROUND(2.0 *eyeRadX*scl);
    cropBox[3] = cropBox[1] + ROUND((eyeUpY+eyeDownY)*scl);
    sprintf(scrName, "%s%s_nose", pathNameOutput, outputFileRoot);
    build_filename(fnOutput, scrName, 3, kTrain, "");
    cropWritePfm(fnOutput, cropBox, warped, nxWarp, nyWarp);

  }  /* Over raw training images */
  fclose(fpIn);
}
