#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <pwd.h>

#define PARGVAL	((*argv)[2] ? (*argv)+2 : --argc ? *++argv : (char *)0)

unsigned char raster[1024 * 3] = { 0 };

int width = 216;		/* picture width in 1/72 inches */
int height = 216;		/* picture height in 1/72 inches */
int x_offset = 36;		/* offset from left edge of page */
int y_offset = 36;		/* offset from bottom of page */
int rasters;			/* number of scanlines in file */
int raster_width;		/* number of pixels in a scanline */
int bits_per_pixel = 8;		/* number of bits per pixel */
int centre;			/* centre the picture in the page */
int flip = 1;			/* flip the picture upside down */
int copies = 1;			/* number of copies */
int headSize = 0;		/* size of header */
int invert;			/* complement the image 1 */
char *printer = "-Plw";		/* printer to spool to */
char *title= NULL;
char *textfile= NULL;

char filnam[40], command[64];

main(argc,argv)
int argc; char **argv;
{
    struct passwd *userentry; int pflag = 1;
    char *username = NULL;
    char *printEnv, *getenv(), string[64];
    FILE *in = stdin, *out = (FILE *)0;
    double atof();
    int kSize;

    if((printEnv= getenv("PRINTER")) != NULL)
	{
	sprintf(string, "-P%s",printEnv);
	printer= string;
	}

    while (--argc > 0) {
	if ((*++argv)[0] == '-')
		switch ((*argv)[1]) {
		case 'P':
		    printer = *argv; break;
		case 'n':
		    pflag = 0; break;
		case 's':
		    out = stdout; break;
		case 'u':
		    username=PARGVAL; break;
		case 'c':
		    centre++; break;
		case 'f':
		    flip = -flip; break;
		case 'T':
		    if (--argc > 0) {
			textfile= argv[1];
			argv++;
		    } else goto usage;
		    break;
		case 't':
		    if (--argc > 0) {
			title= argv[1];
			argv++;
		    } else goto usage;
		    break;
		case 'w':
		    if (--argc > 0) {
			width = atof(argv[1])*72.0;
			argv++;
		    } else goto usage;
		    break;
		case 'h':
		    if (--argc > 0) {
			height = atof(argv[1])*72.0;
			argv++;
		    } else goto usage;
		    break;
		case 'x':
		    if (--argc > 0) {
			raster_width = atoi(argv[1]);
			argv++;
		    } else goto usage;
		    break;
		case 'y':
		    if (--argc > 0) {
			rasters = atoi(argv[1]);
			argv++;
		    } else goto usage;
		    break;
		case 'b':
		    if (--argc > 0) {
			bits_per_pixel = atoi(argv[1]);
			argv++;
		    } else goto usage;
		    break;
		case 'k':
		    if (--argc > 0) {
			copies = atoi(argv[1]);
			/* allow 0 copies for splicing output */
			if (copies < 0) copies = 1;
			argv++;
		    } else goto usage;
		    break;
		case 'o':
		    switch ((*argv)[2]) {
		    case 'x':
			if (--argc > 0) {
			    x_offset = atof(argv[1])*72.0;
			    argv++;
			} else goto usage;
			break;
		    case 'y':
			if (--argc > 0) {
			    y_offset = atof(argv[1])*72.0;
			    argv++;
			} else goto usage;
			break;
		    default: goto usage;
		    }
		    break;

		case 'i':
			invert++;
			break;
		case 'H':
		    if (--argc > 0) {
			headSize = atoi(argv[1]);
			/* allow 0 copies for splicing output */
			if (headSize < 0) headSize = 0;
			argv++;
		    } else goto usage;
		    break;
		default:
		    fprintf(stderr,"halft: unknown option %s\n",*argv);
		    exit(1);
usage:
		    fprintf(stderr, "Usage: halft [-w width] [-h height] [-x x_size] [-y y_size] [-b bits_per_pixel] [-k copies] [-c] [-ox x_offset] [-oy y_offset] [-f] [-H headerSize] [-s] [-Pprinter] [file]\n");
		    exit(1);
	    }
	else {
	    if ((in = fopen(*argv, "r")) == NULL) {
		perror(*argv);
		exit(1);
	    }
	}
    }

    if (out == (FILE *)0) {
	fprintf(stderr,"halft: cannot open output file\n");
	exit(1);
    }

    if (centre) {
	x_offset = (612 - width)/2;
	y_offset = (792 - height)/2;
    }
    if (rasters == 0 || raster_width == 0)
	if (infer(in) == -1) {
	    fprintf(stderr, "halft: unable to determine picture size\n");
	    goto error;
	}
    if (invert && bits_per_pixel != 1) {
	fprintf(stderr, "halft: don't know how to complement with -b > 1\n");
	goto error;
    }

    while(headSize>0) {
     if (headSize > 1024)
      kSize = 1024;
     else
      kSize = headSize;
     if (fread((char *)raster, kSize, 1, in) <= 0) {
	fprintf(stderr,"halft: error reading header in input file\n");
	exit(2);
     }
     headSize -= 1024;
    }

    if (glyphpage(in,out) < 0) {
	fprintf(stderr,"halft: error reading image in input file\n");
	if (out != stdout) {
error:
	    fclose(out);
	    sprintf(command,"rm -f %s &",filnam);
	    system(command);
	}
	exit(2);
    }

    if (copies) {
	fprintf(out, "/#copies %d def\n", copies);
	fputs("showpage\n", out);
    }
    if (out == stdout) exit(0);
    fclose(out);

    /* spool to printer and remove file */
    sprintf(command,"lpr %s -r %s 0</dev/null 1>/dev/null 2>&1 &", printer, filnam);
    if (pflag)
	system(command);
    else
	fprintf(stderr,"halft: %s\n",command);
    exit(0);
}

DoTitle(out)
	FILE *out;
	{
	int myX, myY;

	myX= x_offset+width/2;
	myY= y_offset-72; /* an inch below */

	fputs("\n/Times-Roman findfont\n", out);
	fputs("16 scalefont\n", out);
	fputs("setfont\n", out);
	fprintf(out, "%d %d moveto\n", myX, myY);
	fprintf(out, "(%s)\n", title);
	fputs("dup stringwidth pop \n-", out);
	fputs(".5 mul\n", out);
	fputs("0 rmoveto\n", out);
	fputs("show\n\n", out);
	}

DoText(out)
	FILE *out;
	{
	int myX, myY;
	FILE *text_fp;
	char buf[1024];

	if (! (text_fp=fopen(textfile,"r")) ) {
	    perror(textfile);
	    exit(1);
	    }
	myX= x_offset+width;
	if (myX > 72*(8-3)) {
	    /* too close to page edge */
	    myX = 72*(8-3);
	    }
	myY = 72*10.5; /* half inch from top */

	fputs("\n/Times-Roman findfont\n", out);
	fputs("8 scalefont\n", out);
	fputs("setfont\n", out);
	fprintf(out, "%d %d moveto\n", myX, myY);
	fputs("0.2 setgray\n", out); /* so it shown up on black (a bit) */

	/* compute a width, assume height is about the same */
	fputs("/ystep (mo) stringwidth pop def\n",out);

	for (;fscanf(text_fp,"%[^\n]",buf)>0;) {
	    fgetc(text_fp);

	    /* save current position */
	    fputs("currentpoint\n", out);

	    fprintf(out, "(%s) show \n", buf);

	    /* add y offset to y and go there */
	    fputs("ystep sub moveto \n", out);

	    fputs("\n", out);
	    }
	fputs("0 setgray\n", out);
	fclose(text_fp);
	}

glyphpage(in, out)
FILE *in, *out;
{
    register int krast, i, iprev, j;
    int w;
    int q = 0;

    fputs("%!\n", out);
    if(title)
	DoTitle(out);
    if(textfile) fputs("gsave\n",out);

    fputs("/picstr 36 string def\n", out);
    fprintf(out, "%d %d translate\n", x_offset, y_offset);
    fprintf(out, "%d %d scale\n", width, height);
    fprintf(out, "%d %d %d [%d 0 0 %d 0 %d] { currentfile picstr readhexstring pop } image\n",
	    raster_width, rasters, (bits_per_pixel==24) ? 8 : bits_per_pixel,
	    raster_width, flip*rasters, (flip==1) ? 0 : rasters);
    w = (bits_per_pixel == 24) ? raster_width : (raster_width*bits_per_pixel)/8;
    for (; rasters>0; rasters--) {
	if (readrast(raster,raster_width,in) < 0) return -1;
	for (j = 0; j < w; j++) {
	    if (invert)
		fprintf(out, "%02x", (~raster[j])&0xff);
	    else
		fprintf(out, "%02x", raster[j]&0xff);
	    if (++q == 36) {
		q = 0;
		putc('\n', out);
	    }
	}
    }
    if (q != 0) {
	while(q++ != 36)
	    fprintf(out, "00");
	putc('\n', out);
    }
    if(textfile) {
	fputs("grestore\n",out);
	DoText(out);
    }
    return 0;
}

readrast(p1, width, in)
unsigned char *p1; int width; FILE *in;
{
    if (fread((char *)p1, (width*bits_per_pixel)/8, 1, in) <= 0) return -1;
    if (bits_per_pixel == 24) {
	/* do YIQ conversion */
	register unsigned char *p2, *p3;
	for (p2 = p3 = p1; p3 < p1+width; p3++, p2 += 3)
	    *p3 = (int)(0.3*p2[0] + 0.59*p2[1] + 0.11*p2[2]);
    }
    return 0;
}

infer(in)
FILE *in;
{
    struct stat statb;
    long size, i;

    fstat(fileno(in), &statb);
    size = statb.st_size;
    /* try to infer the size for square images */
    if (bits_per_pixel) {
	for (i = 1; i <= 1024; i++)
	    if (i*i*bits_per_pixel == size*8) {
		rasters = i;
		raster_width = i;
		return 0;
	    }
	return -1;
    } else {
	for (bits_per_pixel = 1;; bits_per_pixel += 1) {
	    for (i = 1; i <= 1024; i++)
		if (i*i*bits_per_pixel == size*8) {
		    rasters = i;
		    raster_width = i;
		    return 0;
		}
	    if (bits_per_pixel == 24) break;
	}
    }
    return -1;
}
