/***  STALE ***/
# include "phase.h"
# include "svd.h"

extern float machineEps;

float linear_interp(low_value, high_value, position)

     float low_value, high_value, position;
{
  float result;

  result = ((low_value * (1 - position))
	    + (high_value * position));
  return(result);
}

/*
Backwards warp of "image" by warp b.  The result is stored in "warped".
Pixel values are determined by linearly interpolating the warped image.
While this is fairly fast, a better interpolation may be desirable.
*/
affine_warp_image(image, warped, x_cent, y_cent, b, nx, ny, defaultGrey)

     float *image, *warped;
     float b[6], x_cent, y_cent;
     int nx, ny;
     unsigned char defaultGrey;
{
  float linear_interp(), new_i, new_j, remainder_i, remainder_j;
  int index, i, j, floor_i, floor_j, ceil_i, ceil_j;
  float u, v;


  for(i=0;i<ny;i++){
    for(j=0;j<nx;j++){
      index = (i*nx) + j;
      new_j = ((float) j) + b[0] + ((j-x_cent) * b[2]) + ((i-y_cent) * b[4]);
      new_i = ((float) i) + b[1] + ((j-x_cent) * b[3]) + ((i-y_cent) * b[5]);

      if (new_j < 0 || new_j > (nx - 1) || new_i < 0 || new_i > (ny - 1)) 
	warped[index] = defaultGrey;
      else{
	floor_i = ((int) floor(new_i));
	ceil_i = ((int) ceil(new_i));
	remainder_i = new_i - ((float) floor_i);
	floor_j = ((int) floor(new_j));
	ceil_j = ((int) ceil(new_j));
	remainder_j = new_j - ((float) floor_j);
      
	warped[index] = 
	  linear_interp
	  (linear_interp(image[(floor_i*nx)+floor_j],
			 image[(ceil_i*nx)+floor_j],
			 remainder_i),
	   linear_interp(image[(floor_i*nx)+ceil_j],
			 image[(ceil_i*nx)+ceil_j],
			 remainder_i),
	   remainder_j);
      }
    }
  }
}

/*
Backwards warp of "image" by warp b.  The result is stored in "warped".
Pixel values are determined by linearly interpolating the warped image.

Here the warped image can be resampled at a different rate
*/
affine_warp_sample_image(image, nx, ny,
                         warped, xMin, yMin, dx, dy, newNx, newNy,
                         x_cent, y_cent, b, defaultGrey)

     float *image, *warped;
     float xMin, yMin, dx, dy;
     float b[6], x_cent, y_cent;
     int nx, ny, newNx, newNy; 
     unsigned char defaultGrey;
{
  float linear_interp(), new_i, new_j, remainder_i, remainder_j;
  int index, i, j, floor_i, floor_j, ceil_i, ceil_j;
  float u, v, yTmp, xTmp;


  for(i=0;i<newNy;i++){
    yTmp = yMin + dy*i;
    for(j=0;j<newNx;j++){
      xTmp = xMin + dx*j;
      index = (i*newNx) + j;
      new_j = xTmp + b[0] + ((xTmp-x_cent) * b[2]) + ((yTmp-y_cent) * b[4]);
      new_i = yTmp + b[1] + ((xTmp-x_cent) * b[3]) + ((yTmp-y_cent) * b[5]);

      if (new_j < 0 || new_j > (nx - 1) || new_i < 0 || new_i > (ny - 1)) 
	warped[index] = defaultGrey;
      else{
	floor_i = ((int) floor(new_i));
	ceil_i = ((int) ceil(new_i));
	remainder_i = new_i - ((float) floor_i);
	floor_j = ((int) floor(new_j));
	ceil_j = ((int) ceil(new_j));
	remainder_j = new_j - ((float) floor_j);
      
	warped[index] = 
	  linear_interp
	  (linear_interp(image[(floor_i*nx)+floor_j],
			 image[(ceil_i*nx)+floor_j],
			 remainder_i),
	   linear_interp(image[(floor_i*nx)+ceil_j],
			 image[(ceil_i*nx)+ceil_j],
			 remainder_i),
	   remainder_j);
      }
    }
  }
}


/** Compute Inverse Affine Image Transform :

Definition:
    Affine Transform:
      x_new = x_cent + a + (I + A) (x_old - x_cent)

    with: 
      x_old - a 2 vector,
      x_new - the warped coordinats
      x_cent = center point.
      a and A the six affine parameters.

    Inverse Affine Transform:

     x_old = x_cent + b + (I + B) (x_new - x_cent)

    with
     B = inverse(I+A) - I
     b = -(I+B) * a

Subroutine:
     int invertAffine(a, b)
   
     External:
      float  machineEps = 1.0e-6;
      # define        abs(A)        ((A)<0?-(A):(A))

     Input:  Float affine parameters
      a = ( a_0, a_1, A_00, A_10, A_01, A_11)
      (NOTE: Order is down columns first)

     Returns:
      0 if successful
     -1 if failure due to singularity of input transformation.

     Output:  Float affine parameters
      b = ( b_0, b_1, B_00, B_10, B_01, B_11)
     (NOTE: Order is down columns first)

**************************************************************/

int invertAffine(a, b)
 float a[6], b[6];
 {
  float det;

  /* Compute determinant( I+A ) where:
                   | a[2]    a[4] |
      I + A  = I + |              |
                   | a[3]    a[5] |
  */
  det = (1.0 + a[2]) * (1.0 + a[5]) - a[3] * a[4];
  if (abs(det) < machineEps) 
   return((int)-1);

  /* set B = inverse(I+A) */
  b[2] = (1.0 + a[5])/det;  /* Switch diagonal elements of (I+A)/det */
  b[5] = (1.0 + a[2])/det;
  b[3] = -a[3]/det;         /* Divide off-diagonal elements by -det */
  b[4] = -a[4]/det;
  /*  now matrix represented by b[2-5] is I+B == inverse(I+A) */
 
  /* Compute b0, b1 according to b = -inverse(I+A) a */
  b[0] = -b[2] * a[0] - b[4] * a[1];
  b[1] = -b[3] * a[0] - b[5] * a[1];
  
  /* Revise b[2-5] to represent B (not I+B, as it does now) */
  b[2] -= 1.0;
  b[5] -= 1.0;
  fprintf(stderr," b= %e %e\n    %e %e %e %e\n", b[0], b[1],
    b[2], b[3], b[4], b[5]);

  return((int) 0);
 }

 
updateAffine(ixA,iyA, a, ixB, iyB, b)
  float a[6], b[6];
  float ixA, iyA, ixB, iyB;
  {
    /* Evaluate composition of affine transforms a <- b o a */
    /* Center position remains ixA, iyA */

    float scr[6];
    int i;

    /* a'' = a + b + B * ( a + iA - iB) */
    scr[0] = a[0] + b[0] + b[2] * (a[0]+(ixA-ixB)) + b[4] *(a[1]+(iyA-iyB));
    scr[1] = a[1] + b[1] + b[3] * (a[0]+(ixA-ixB)) + b[5] *(a[1]+(iyA-iyB));

    /* A'' = A + B + B * A    */
    scr[2] = a[2] + b[2] + b[2] * a[2] + b[4] * a[3];
    scr[3] = a[3] + b[3] + b[3] * a[2] + b[5] * a[3];
    scr[4] = a[4] + b[4] + b[2] * a[4] + b[4] * a[5];
    scr[5] = a[5] + b[5] + b[3] * a[4] + b[5] * a[5];

    for(i=0; i<6; i++)
     a[i] = scr[i];
 
 }

/** Compute Sample Value of Affine Image Transform */

affineSample(x, y, aff, v)
 float x,y,aff[6];
 float v[2];
 {
  v[0] = aff[0] + aff[2] * x + aff[4] * y ;
  v[1] = aff[1] + aff[3] * x + aff[5] * y ;
 }


affinePtSample(x, y, ixcent, iycent, aff, v)
 float x,y,aff[6];
 float v[2];
 float ixcent, iycent;
 {
  v[0] = x + aff[0] + aff[2] * (x - ixcent) + aff[4] * (y -iycent);
  v[1] = y + aff[1] + aff[3] * (x - ixcent) + aff[5] * (y -iycent);
 }


/** Compute Sample Value of Rational Image Transform */

int rationalSample(x, y, x0, y0, aff, v)
/*  Sample rational transform:
      v-x0_vec = (a + A(x_vec-x0_vec)/(1+b^T (x_vec-x0_vec))
    where x0_vec = origin for transform.  Here
    x_vec = (x, y)^T, and similarly for x0_vec.
***************************************************/
    
 float x,y,x0,y0,aff[8];
 float v[2];
 {
  float scl;
  
  x -= x0;
  y -= y0;
  scl = 1.0 + aff[6]*x + aff[7]*y;

  if (scl < -machineEps) { /* Invalid argument, denominator < zero */ 
   v[0] = (aff[0] + aff[2] * x + aff[4] * y)/scl + x0 ;
   v[1] = (aff[1] + aff[3] * x + aff[5] * y)/scl + y0;
   return((int) -2);
  } else if (scl < machineEps) { /* Invalid argument, denominator apprx= zero */
   v[0] = v[1] = 0.0;
   return((int) -1);
  } else {
   v[0] = (aff[0] + aff[2] * x + aff[4] * y)/scl + x0 ;
   v[1] = (aff[1] + aff[3] * x + aff[5] * y)/scl + y0;
   return((int) 0);
  }
 }

int invertRational(a, b)
 float a[8], b[8];
/*  Given rational transform:
      x'-x0 = (a + A(x-x0)/(1+b^T (x-x0))
    where x0 = origin for transform,
    compute inverse
      x-x0 = (a' + A'(x-x0))/(1+b'^T (x-x0))
    
    Here the original transform is given in terms of

    a[0-7] = (a_1 a_2 A_11 A_21 A_12 A_22 b_1 b_2)
    indices.. 0   1   2    3    4    5    6   7

    and similarly, the inverse transform is returned
    in  b[0-7]

    Returns value 0 if successful, else -1 if transform is singular
*/
 {
  float det;

  /* Compute determinant( A ) where:
            | a[2]    a[4] |
      A  =  |              |
            | a[3]    a[5] |
  */
  det = a[2] * a[5] - a[3] * a[4];
  if (abs(det) < machineEps) 
   return((int)-1);

  /* set B = inverse(A) */
  b[2] = a[5]/det;  /* Switch diagonal elements of (I+A)/det */
  b[5] = a[2]/det;
  b[3] = -a[3]/det;         /* Divide off-diagonal elements by -det */
  b[4] = -a[4]/det;
  /*  now matrix represented by b[2-5] is B == inverse(A) */
 
  /* Compute b' according to b' = -inverse(A)^T b */
  b[6] = -(b[2] * a[6] + b[3] * a[7]);
  b[7] = -(b[4] * a[6] + b[5] * a[7]);

  /* Compute a' according to a' = -inverse(A) a */
  b[0] = -b[2] * a[0] - b[4] * a[1];
  b[1] = -b[3] * a[0] - b[5] * a[1];
  
  /* Revise b[2-5] to represent A' = inverse(A) - ( b_2) (a_2 , -a_1) /det(A)
                                                  (-b_1)                     */
  b[2] -= a[7] * a[1] /det;
  b[3] -= -a[6] * a[1] /det;
  b[4] -= -a[7] * a[0] /det;
  b[5] -= a[6] * a[0] /det;

  return((int) 0);
 }

 
updateRational(ixR,iyR, rFlt, ixD, iyD, dFlt)
  float rFlt[8], dFlt[8];
  float ixR, iyR, ixD, iyD;
  {
    /* Evaluate composition of rational transforms r <- d o r */
    /* Center position remains ixR, iyR */

    double r[8], d[8];
    double scr[8], w, tmp, s[2], Rprime[2][2], Atmp[2][2];
    int i;

    for(i=0; i<8; i++) {
     r[i] = (double) rFlt[i];
     d[i] = (double) dFlt[i];
    }

    /* a'' =  a_d + A_d * (a_r+c_r-c_d) + (c_d-c_r){1+b_d^T[a_r+c_r-c_d]} */

    scr[0] = d[0] + d[2] * (r[0]+ixR-ixD) + d[4] * (r[1]+iyR-iyD);
    scr[1] = d[1] + d[3] * (r[0]+ixR-ixD) + d[5] * (r[1]+iyR-iyD);
    tmp = 1 + d[6] * ( r[0] + ixR - ixD) + d[7] * ( r[1] + iyR - iyD);
    scr[0] += (ixD-ixR) * tmp;
    scr[1] += (iyD-iyR) * tmp;

    /* b'' = b_r^T + b_d dot R'
       with R' = (A_r + (c_r-c_d) b_r^T) */
    Rprime[0][0] = r[2] + (ixR -ixD) * r[6];
    Rprime[0][1] = r[4] + (ixR -ixD) * r[7];
    Rprime[1][0] = r[3] + (iyR -iyD) * r[6];
    Rprime[1][1] = r[5] + (iyR -iyD) * r[7];

    /* b'' = b_r^T + b_d dot R' */
    scr[6] = r[6] + d[6] * Rprime[0][0] + d[7] * Rprime[1][0] ;
    scr[7] = r[7] + d[6] * Rprime[0][1] + d[7] * Rprime[1][1] ;

    /* A'' = (a_d+c_d-c_r ) b_r^T + (A_d + (c_d-c_r) b_d^T) Rprime */

    Atmp[0][0] = d[2] + (ixD-ixR) * d[6];
    Atmp[0][1] = d[4] + (ixD-ixR) * d[7];
    Atmp[1][0] = d[3] + (iyD-iyR) * d[6];
    Atmp[1][1] = d[5] + (iyD-iyR) * d[7];

    scr[2] = Atmp[0][0] * Rprime[0][0] + Atmp[0][1] * Rprime[1][0];
    scr[3] = Atmp[1][0] * Rprime[0][0] + Atmp[1][1] * Rprime[1][0];
    scr[4] = Atmp[0][0] * Rprime[0][1] + Atmp[0][1] * Rprime[1][1];
    scr[5] = Atmp[1][0] * Rprime[0][1] + Atmp[1][1] * Rprime[1][1];

    scr[2] += (d[0] + ixD - ixR) * r[6];
    scr[4] += (d[0] + ixD - ixR) * r[7];
    scr[3] += (d[1] + iyD - iyR) * r[6];
    scr[5] += (d[1] + iyD - iyR) * r[7];

    /* Rescale by w=1/b0, b0 = new constant term = 1+d_b dot (a_r + c_r - c_d) */

    w = 1.0/(1.0 + d[6] * (r[0]+ixR-ixD) + d[7] * (r[1]+iyR-iyD));
    for(i=0; i<8; i++)
     r[i] = w * scr[i];
 
    for(i=0; i<8; i++) {
     rFlt[i] = (float) r[i];
    }
 }

/* move the center of the rational map a from x0,y0 to x1,y1 */
moveRationalCenter(x0,y0, a, x1,y1)
 float x0,y0, a[8], x1, y1;
 {
  float b[8];
  int i;
  /* Make identity map */
  for(i=0; i<8; i++)
   b[i] = 0.0;
  b[2] = b[5] = 1.0;
  updateRational(x1,y1, b, x0,y0, a);
  for(i=0; i<8; i++)
   a[i] = b[i];
 }
  
/* Map orientation vector, inVec, at x,,y to new orientation
   according to mapping a */
mapRationalVec(inVec, x, y, x0, y0, a, outVec)
 float inVec[2], x, y, x0,y0, a[8], outVec[2];
 {
  float tail[2], head[2], nrm;
 
  nrm = inVec[0] * inVec[0] + inVec[1] * inVec[1];
  if (nrm > 0.0) nrm = sqrt((double) nrm);
  else nrm = 0.0;
  
  rationalSample(x, y, x0, y0, a, tail);
  rationalSample(x+inVec[0]/nrm, y+inVec[1]/nrm, x0, y0, a, head);

  outVec[0] = (head[0] - tail[0] ) * nrm;
  outVec[1] = (head[1] - tail[1] ) * nrm;
  
 }

/*
Backwards warp of "image" by rational warp b.  The result is stored in "warped".
Pixel values are determined by linearly interpolating the warped image.
While this is fairly fast, a better interpolation may be desirable.
*/
rational_warp_image(image, warped, x_cent, y_cent, b, nx, ny, defaultGrey)

     float *image, *warped;
     float b[8], x_cent, y_cent;
     int nx, ny;
     unsigned char defaultGrey;
{
  float linear_interp(), new_i, new_j, remainder_i, remainder_j;
  float u, v, scr[2];

  int index, i, j, floor_i, floor_j, ceil_i, ceil_j;


  for(i=0;i<ny;i++){
    for(j=0;j<nx;j++){
      index = (i*nx) + j;

      if (rationalSample((float)j,(float)i, x_cent,y_cent,b, scr) >= 0) {
       new_j = scr[0];
       new_i = scr[1];
      } else
       new_j = -1; /* rational map invalid */

      if (new_j < 0 || new_j > (nx - 1) || new_i < 0 || new_i > (ny - 1)) 
	warped[index] = defaultGrey;
      else{
	floor_i = ((int) floor(new_i));
	ceil_i = ((int) ceil(new_i));
	remainder_i = new_i - ((float) floor_i);
	floor_j = ((int) floor(new_j));
	ceil_j = ((int) ceil(new_j));
	remainder_j = new_j - ((float) floor_j);
      
	warped[index] = 
	  linear_interp
	  (linear_interp(image[(floor_i*nx)+floor_j],
			 image[(ceil_i*nx)+floor_j],
			 remainder_i),
	   linear_interp(image[(floor_i*nx)+ceil_j],
			 image[(ceil_i*nx)+ceil_j],
			 remainder_i),
	   remainder_j);
      }
    }
  }
}

scoreBox( image, box, nx, ny)

 float *image;
 int nx, ny, *box;
{
  int index, i, j, i0, j0, i1, j1; 

  i0=box[0]; j0 = box[1];
  i1=box[2]; j1 = box[3];
  fprintf(stderr, "debug: drawing box %d %d %d %d\n", i0, j0, i1, j1);

  if (j0>=0 && j0 < ny) {
   /* Draw Top */
   i = max(0, i0);
   for(;i<min(nx,i1);i++) {
     index = i + j0 * nx;
     if (image[index] < 128.0)
      image[index] = 255.0;
     else
      image[index] = 0.0;
    }
   }

  if (j1>=0 && j1 < ny) {
   /* Draw Bottom */
   i = max(0, i0);
   for(;i<min(nx,i1);i++) {
     index = i + j1 * nx;
     if (image[index] < 128.0)
      image[index] = 255.0;
     else
      image[index] = 0.0;
    }
   }
  
  if (i0>=0 && i0 < nx) {
   /* Draw Left */
   j = max(0, j0);
   for(;j<min(ny,j1);j++) {
     index = i0 + j * nx;
     if (image[index] < 128.0)
      image[index] = 255.0;
     else
      image[index] = 0.0;
    }
   }
  
  if (i1>=0 && i1 < nx) {
   /* Draw Right */
   j = max(0, j0);
   for(;j<min(ny,j1);j++) {
     index = i1 + j * nx;
     if (image[index] < 128.0)
      image[index] = 255.0;
     else
      image[index] = 0.0;
    }
   }
}

/*
Backwards warp of "image" according to radial warp.
The result is stored in "warped".
Pixel values are determined by linearly interpolating the warped image.
While this is fairly fast, a better interpolation may be desirable.
*/
radial_warp_image(image, warped, par, nx, ny, defaultGrey)

     float *image, *warped;
     float par[5];
     int nx, ny;
     unsigned char defaultGrey;
{
  float linear_interp(), new_i, new_j, remainder_i, remainder_j;
  float u, v, scr[2];
  int index, i, j, floor_i, floor_j, ceil_i, ceil_j;


  for(i=0;i<ny;i++){
    for(j=0;j<nx;j++){
      index = (i*nx) + j;

      if (inverseRadial((float)j,(float)i, par[0],par[1],par+2, scr) >= 0) {
       new_j = scr[0];
       new_i = scr[1];
      } else
       new_j = -1; /* inverse map invalid */

      if (new_j < 0 || new_j > (nx - 1) || new_i < 0 || new_i > (ny - 1)) 
	warped[index] = defaultGrey;
      else{
	floor_i = ((int) floor(new_i));
	ceil_i = ((int) ceil(new_i));
	remainder_i = new_i - ((float) floor_i);
	floor_j = ((int) floor(new_j));
	ceil_j = ((int) ceil(new_j));
	remainder_j = new_j - ((float) floor_j);
      
	warped[index] = 
	  linear_interp
	  (linear_interp(image[(floor_i*nx)+floor_j],
			 image[(ceil_i*nx)+floor_j],
			 remainder_i),
	   linear_interp(image[(floor_i*nx)+ceil_j],
			 image[(ceil_i*nx)+ceil_j],
			 remainder_i),
	   remainder_j);
      }
    }
  }
}

/* Invert the radial distortion map
  x_vec = (w - x0_vec) * (1 + a0 * ||w - x0_vec||^2) + x0_vec
to find w */

int inverseRadial(x,y, x0,y0, a, w)
 float x,y, x0,y0, a[3], w[2];
 {
   float r, r2, rw, res;
   int i;
  
   r2 = (x-x0)*(x-x0) + (y-y0)*(y-y0);
   r = sqrt(r2);
   
   /* radius rw satisfies rw * (1 + a0 *rw^2) - r = 0 */
   if (r < machineEps) {
    rw = 0.0;
    w[0] = x0;
    w[1] = y0;
   } else {
    rw = r/(1+ a[0] * r2);
    for(i=0; i< 10; i++) {
     res = rw * (1+ a[0] * rw * rw) - r;
     if (abs(res)/r <= machineEps) 
      break;
     rw -= res/( 1 + 3.0 * a[0] * rw * rw); /* Newton step */
    }
    /*** DEBUG **/
    if (i>5)
     fprintf(stderr, "InverseWarp: x,y = %e %e takes %d its\n",x,y,i );

    /* Compute corrected point */
    w[0] = x0 + rw/r * (x - x0);
    w[1] = y0 + rw/r * (y - y0);
   }
 }

warpAndBound(image, nx, ny, pwarped, pnxw, pnyw, x_cent, y_cent, a, defaultGrey,
                subSample, b)
  /*
  Forwards warp of "image" by rational warp a.
  The result is stored in "warped".
  The size of warped is the smallest upright rectangle containing
  the mapping of "image".  The bounding box is restricted to lie on
  a subsampled grid, starting at 0,0 with spacing `subSample'.
  The size of warped is returned as *pnxw, *pnyw.
  Grey levels are determined by linearly interpolating the warped image.
  While this is fairly fast, a better interpolation may be desirable.

  Returns the backwards warp from new warped image coords to original
  image in b with center x_cent, y_cent left unchanged 
  */

     float *image, **pwarped;
     float a[8], b[8], x_cent, y_cent;
   
     int nx, ny, *pnxw, *pnyw, subSample;
     unsigned char defaultGrey;
{
  float linear_interp(), new_i, new_j, remainder_i, remainder_j;
  float u, v, scr[2], *warp;
  float c[8], boxCorners[2][2], imageVertex[4][2];

  int index, i, j, floor_i, floor_j, ceil_i, ceil_j;
  int ix0, iy0, ix1, iy1, size;

  /* compute inverse mapping (check for singularity) */
  if (-1 == invertRational(a, c))
    error("Rational transform for prewarp nearly singular","");
  /* c maps from warped coords to original "image" coords */

  /* Computed bounding upright rectangle */

  /* Compute box corners */
  rationalSample((float)0.0, (float)0.0, x_cent, y_cent, a, imageVertex[0]);
  rationalSample((float)0.0, (float)(ny-1), x_cent, y_cent, a, imageVertex[1]);
  rationalSample((float)(nx-1), (float)(ny-1), x_cent, y_cent, a, imageVertex[2]);
  rationalSample((float)(nx-1), (float)0.0, x_cent, y_cent, a, imageVertex[3]);

  boxCorners[0][0] = imageVertex[0][0];
  for(j=1; j<4; j++)
        boxCorners[0][0] = min(boxCorners[0][0],  imageVertex[j][0]);
  boxCorners[0][1] = imageVertex[0][1];
  for(j=1; j<4; j++)
        boxCorners[0][1] = min(boxCorners[0][1],  imageVertex[j][1]);
  boxCorners[1][0] = imageVertex[0][0];
  for(j=1; j<4; j++)
        boxCorners[1][0] = max(boxCorners[1][0],  imageVertex[j][0]);
  boxCorners[1][1] = imageVertex[0][1];
  for(j=1; j<4; j++)
        boxCorners[1][1] = max(boxCorners[1][1],  imageVertex[j][1]);
  /* Choose box placement to match subSampled grid */
  if (subSample <= 0) subSample = 1; 
  ix0 = floor(boxCorners[0][0]/subSample) * subSample;
  iy0 = floor(boxCorners[0][1]/subSample) * subSample;
  ix1 = ceil(boxCorners[1][0]/subSample) * subSample;
  iy1 = ceil(boxCorners[1][1]/subSample) * subSample;

  *pnxw = ix1 - ix0;
  *pnyw = iy1 - iy0;
  size = (*pnxw) * (*pnyw) ;
  fprintf(stderr," Warped template size %d %d\n", *pnxw, *pnyw);
  grabFloatMemory(pwarped, size, "warped image");
  warp = *pwarped;

  /* Modify backwards warp c according to the translation
     of coords which takes top left corner of warped image,
     in canonical coords, namely (0, 0), to pixel position
     (ix0, iy0) */
  for(i=0; i<8; i++)
   b[i] = 0.0;
  b[2] = b[5] = 1.0;
  b[0] = (float) ix0;
  b[1] = (float) iy0;
  /* b <= c o b */
  updateRational(x_cent,y_cent, b,  x_cent, y_cent, c);

  for(i=0;i<(*pnyw);i++){
    for(j=0;j<(*pnxw);j++){
      index = i * (*pnxw) + j;

      /* Map backwards from warp pixel i,j, to orig image pixel scr[*] */
      if (rationalSample((float)j,(float)i, x_cent,y_cent,b, scr) >= 0) {
       new_j = scr[0];
       new_i = scr[1];
      } else
       new_j = -1; /* rational map invalid */

      if (new_j < 0 || new_j > nx - 1 || new_i < 0 || new_i > ny - 1) 
	warp[index] = defaultGrey;
      else{
	floor_i = ((int) floor(new_i));
	ceil_i = ((int) ceil(new_i));
	remainder_i = new_i - ((float) floor_i);
	floor_j = ((int) floor(new_j));
	ceil_j = ((int) ceil(new_j));
	remainder_j = new_j - ((float) floor_j);
	warp[index] = 
	  linear_interp
	  (linear_interp(image[(floor_i*nx)+floor_j],
			 image[(ceil_i*nx)+floor_j],
			 remainder_i),
	   linear_interp(image[(floor_i*nx)+ceil_j],
			 image[(ceil_i*nx)+ceil_j],
			 remainder_i),
	   remainder_j);
      }
    }
  }

}

warpCrop(image, nx, ny, pwarped, ix0, iy0, ix1, iy1, x_cent, y_cent,
          a, defaultGrey)
  /*
  Forwards warp of "image" by rational warp a.
  The result is stored in "warped".

  The size of warped cropped to bounding box with top left
  corner ix0,iy0, and bottom right corner ixy, iy1.
  Grey levels are determined by linearly interpolating the warped image.
  While this is fairly fast, a better interpolation may be desirable.

  */

     float *image, **pwarped;
     float a[8], x_cent, y_cent;
   
     int nx, ny, ix0, iy0, ix1, iy1;
     unsigned char defaultGrey;
{
  float linear_interp(), new_i, new_j, remainder_i, remainder_j;
  float u, v, scr[2], *warp;
  float b[8], c[8];

  int index, i, j, floor_i, floor_j, ceil_i, ceil_j;
  int nxw, nyw, size;

  /* compute inverse mapping (check for singularity) */
  if (-1 == invertRational(a, c))
    error("Rational transform for prewarp nearly singular","");
  /* c maps from warped coords to original "image" coords */

  /* Computed bounding upright rectangle */

  nxw = ix1 - ix0;
  nyw = iy1 - iy0;
  size = nxw * nyw;
  fprintf(stderr," Warped template size %d %d\n", nxw, nyw);
  grabFloatMemory(pwarped, size, "warped image");
  warp = *pwarped;

  /* Modify backwards warp c according to the translation
     of coords which takes top left corner of warped image,
     in canonical coords, namely (0, 0), to pixel position
     (ix0, iy0) */
  for(i=0; i<8; i++)
   b[i] = 0.0;
  b[2] = b[5] = 1.0;
  b[0] = (float) ix0;
  b[1] = (float) iy0;
  /* b <= c o b */
  updateRational(x_cent,y_cent, b,  x_cent, y_cent, c);

  fprintf(stderr,"warpCrop params:\n");
  fprintf(stderr,"%6.1f %6.1f\n", x_cent, y_cent);
  fprintf(stderr,"%e %e\n", b[0], b[1]);
  fprintf(stderr,"%e %e %e %e\n", b[2], b[3], b[4], b[5]);
  fprintf(stderr,"%e %e\n", b[6], b[7]);

  for(i=0;i<nyw;i++){
    for(j=0;j<nxw;j++){
      index = i * nxw + j;

      /* Map backwards from warp pixel i,j, to orig image pixel scr[*] */
      if (rationalSample((float)j,(float)i, x_cent,y_cent,b, scr) >= 0) {
       new_j = scr[0];
       new_i = scr[1];
      } else
       new_j = -1; /* rational map invalid */

      if (new_j < 0 || new_j > nx - 1 || new_i < 0 || new_i > ny - 1) 
	warp[index] = defaultGrey;
      else{
	floor_i = ((int) floor(new_i));
	ceil_i = ((int) ceil(new_i));
	remainder_i = new_i - ((float) floor_i);
	floor_j = ((int) floor(new_j));
	ceil_j = ((int) ceil(new_j));
	remainder_j = new_j - ((float) floor_j);
	warp[index] = 
	  linear_interp
	  (linear_interp(image[(floor_i*nx)+floor_j],
			 image[(ceil_i*nx)+floor_j],
			 remainder_i),
	   linear_interp(image[(floor_i*nx)+ceil_j],
			 image[(ceil_i*nx)+ceil_j],
			 remainder_i),
	   remainder_j);
      }
    }
  }

}

fitParams(T, ixC, iyC, nSys, tPx, tPy, iPx, iPy, nPair)
 float T[], ixC, iyC; 
 int tPx[], tPy[], iPx[], iPy[];
 int nSys, nPair;
 {
  float w[8], A[8][8];
  float iMean[2], tMean[2];
  float detA, maxA, res, rPnt[2], sumRes;
  int i,j, k;

  /* Generate initial guess for forward warp */
  for(i=0; i<8; i++)
   T[i] = 0.0;
  T[2] = T[5] = 1.0;
  iMean[0] = iMean[1] = 0.0;
  tMean[0] = tMean[1] = 0.0;
  for(i=0; i<nPair; i++) {
   iMean[0] += iPx[i];
   iMean[1] += iPy[i];
   tMean[0] += tPx[i];
   tMean[1] += tPy[i];
  }
  iMean[0] /= nPair;
  iMean[1] /= nPair;
  tMean[0] /= nPair;
  tMean[1] /= nPair;

  T[0] = iMean[0] - tMean[0];
  T[1] = iMean[1] - tMean[1];

  if (nSys > 2) {

   /* Estimate scale and rotation */
   for(j=0; j<3; j++)
    for(i=0; i<3; i++)
     A[i][j] = 0.0;
   for(k=0; k<nPair; k++) {
    /* Constraint from x-mapping */
    w[0] = tPx[k] - tMean[0];
    w[1] = -(tPy[k] - tMean[1]);
    w[2] = -(iPx[k] - iMean[0]);
    for(j=0; j<3; j++)
     for(i=0; i<3; i++)
      A[i][j] += w[i] * w[j];
    /* Constraint from y-mapping */
    w[0] = tPy[k] - tMean[1];
    w[1] = tPx[k] - tMean[0];
    w[2] = -(iPy[k] - iMean[1]);
    for(j=0; j<3; j++)
     for(i=0; i<3; i++)
      A[i][j] += w[i] * w[j];
   }
   /* Now scale and rotation matrix
         (a11, -a21)
         (a21,  a11)
      must satisfy:
          A (a11, a21, 1)^T = (0, 0, res)^T
   *****/
     
   detA = A[0][0] * A[1][1] - A[1][0] * A[0][1];
   maxA = max(A[0][0], A[1][1]);
   maxA = max(maxA, A[0][1]);
   if ( detA/ maxA < machineEps) 
    error(" Alignment data nearly singular\n","");

   /* Apply inverse of 2x2 minor of A to -third col of A */
   T[2] = -( A[1][1] * A[0][2] - A[0][1] * A[1][2])/detA;
   T[3] = -(-A[1][0] * A[0][2] + A[0][0] * A[1][2])/detA;
   T[4] = -T[3];
   T[5] = T[2];
  } 

  /* Convert mapping to std form, with center postion tMean */
  moveRationalCenter(tMean[0], tMean[1], T, (float)ixC, (float) iyC);

/* DEBUG: Check residual */
  fprintf(stderr,"Mapping:\n");
  fprintf(stderr," ixC, iyC = %f %f\n", ixC, iyC);
  fprintf(stderr," T[0], T[1] : %e %e\n", T[0], T[1]);
  fprintf(stderr," T[2] - T[5] : %e %e %e %e\n", T[2], T[3], T[4], T[5]);
  fprintf(stderr," T[6], T[7] : %e %e\n", T[6], T[6]);
  sumRes = 0.0;
  fprintf(stderr,"align:\n  tPx, tPy, iPx, iPy: mPx, mPy: res\n");
  for(k=0; k< nPair; k++) {
   rationalSample((float)tPx[k], (float)tPy[k], (float)ixC, (float)iyC, T, rPnt);
   res = (iPx[k] - rPnt[0]) *  (iPx[k] - rPnt[0]);
   res += (iPy[k] - rPnt[1]) *  (iPy[k] - rPnt[1]);
   fprintf(stderr,"%d %d %d %d : ", tPx[k], tPy[k], iPx[k], iPy[k]);
   fprintf(stderr,"%e %e : %e\n", rPnt[0], rPnt[1], sqrt(res));
   sumRes += res;
  }
  fprintf(stderr," RMS = %e\n", sqrt(sumRes)/nPair);

 }

fitParams2(T, ixC, iyC, nSys, tPx, tPy, iPx, iPy, nPair)
 float T[], ixC, iyC;
 int tPx[], tPy[], iPx[], iPy[];
 int nSys, nPair;
 {
  double w[SVD_DIM], A[SVD_DIM][SVD_DIM], b[SVD_DIM], deltaT[SVD_DIM];
  double e[2], gradA[2][8], gradD[8];
  float tmp, res, rPnt[2], sumRes, c;
  float s, maxS, minS, newT[8], newRes;
  int i,j, k,l, nSys0, iIts;

  /* First fit affine model */
  nSys0 = min(nSys, 6);
  for(i=0; i<=8; i++) {
    T[i] = 0.0;
    b[i] = 0.0;
    for(j=0; j<=8; j++)
     A[i][j] = 0.0;
  }
  T[2] = T[5] = 1.0;

  /* Find least squares solution for imagePx,y[*] = affine(trainingPx,y[*]) */
  for(k=0; k< nPair; k++) { /* Build constraint matrix */

   /* Initial error, using identity initial transform */
   e[0] = tPx[k]-iPx[k];
   e[1] = tPy[k]-iPy[k];

   /* Gradient of affine mapping */
   for(j=0; j<2; j++)
    for(i=0; i<8; i++)
     gradA[j][i] = 0.0;
   gradA[0][0] = 1.0;
   gradA[1][1] = 1.0;
   gradA[0][2] = tPx[k];
   gradA[1][3] = tPx[k];
   gradA[0][4] = tPy[k];
   gradA[1][5] = tPy[k];

   /* update least squares constraint matrix */
   for(l =0; l<2; l++)
    for(i=0; i<8; i++) {
     b[i] += e[l] * gradA[l][i];
     for(j=0; j<8; j++)
      A[i+1][j+1] += gradA[l][i] * gradA[l][j];
    }

  } /* over training points: k */

  solveLinearSys(A, deltaT, b, nSys0, w);
  for(i=0; i<nSys0; i++)
    T[i] -= deltaT[i];
  if (nSys < 8) {
   for(i=nSys0; i<8; i++)
    T[i] = 0.0;
  }
 
  if (nSys > 6) {
   sumRes = 0.0;
   fprintf(stderr,"Affine Align:\n  tPx, tPy, iPx, iPy: mPx, mPy: res\n");
   for(k=0; k< nPair; k++) {
    rationalSample((float)tPx[k], (float)tPy[k], (float)0.0, (float)0.0, T, rPnt);
    res = (iPx[k] - rPnt[0]) *  (iPx[k] - rPnt[0]);
    res += (iPy[k] - rPnt[1]) *  (iPy[k] - rPnt[1]);
    fprintf(stderr,"%d %d %d %d : ", tPx[k], tPy[k], iPx[k], iPy[k]);
    fprintf(stderr,"%e %e : %e\n", rPnt[0], rPnt[1], sqrt(res));
    sumRes += res;
   }
   fprintf(stderr," affine RMS = %e\n", sqrt(sumRes)/nPair);

   /* Now have an affine model.  Do gradient descent to find
     rational model */
   for(iIts=0; iIts<50; iIts++) {

   for(i=0; i<=8; i++) {
    b[i] = 0.0;
    for(j=0; j<=8; j++)
     A[i][j] = 0.0;
   }
   for(k=0; k<nPair; k++) {
    rationalSample((float)tPx[k], (float)tPy[k], (float)0.0, (float)0.0, T, rPnt);

    /* Current error */
    e[0] = rPnt[0]-iPx[k];
    e[1] = rPnt[1]-iPy[k];

    /* Denominator of rational map */
    tmp = 1.0 + T[6] * tPx[k] + T[7] * tPy[k];

    /* Gradient of mapping */
    for(j=0; j<2; j++)
     for(i=0; i<8; i++)
      gradA[j][i] = 0.0;
    gradA[0][0] = 1.0/tmp;
    gradA[1][1] = 1.0/tmp;
    gradA[0][2] = tPx[k]/tmp;
    gradA[1][3] = tPx[k]/tmp;
    gradA[0][4] = tPy[k]/tmp;
    gradA[1][5] = tPy[k]/tmp;
    gradA[0][6] = -tPx[k] * rPnt[0]/tmp;
    gradA[0][7] = -tPy[k] * rPnt[0]/tmp;
    gradA[1][6] = -tPx[k] * rPnt[1]/tmp;
    gradA[1][7] = -tPy[k] * rPnt[1]/tmp;

    /* update least squares constraint matrix */
    for(l =0; l<2; l++)
     for(i=0; i<8; i++) {
      b[i] += e[l] * gradA[l][i];
      for(j=0; j<8; j++)
       A[i+1][j+1] += gradA[l][i] * gradA[l][j];
     }

    /* Second derivative terms */

    /* Terms from second derivative of denominator */
    c = 2.0 * (e[0] * rPnt[0] + e[1] * rPnt[1]) / (tmp * tmp);
    A[7][7] += c * tPx[k] * tPx[k];
    A[7][8] += c * tPx[k] * tPy[k];
    A[8][7] += c * tPy[k] * tPx[k];
    A[8][8] += c * tPy[k] * tPy[k];
    
    /* Cross terms */

    /* Gradient of AFFINE mapping */
    for(j=0; j<2; j++)
     for(i=0; i<8; i++)
      gradA[j][i] = 0.0;
    gradA[0][0] = 1.0;
    gradA[1][1] = 1.0;
    gradA[0][2] = tPx[k];
    gradA[1][3] = tPx[k];
    gradA[0][4] = tPy[k];
    gradA[1][5] = tPy[k];

    /* Gradient of denominator alone */
    for(i=0; i<8; i++)
      gradD[i] = 0.0;
    gradD[6] = -tPx[k]/(tmp*tmp);
    gradD[7] = -tPy[k]/(tmp*tmp);
    for(l =0; l<2; l++) {
     for(i=0; i<8; i++) 
      for(j=0; j<8; j++)
       A[i+1][j+1] += e[l] * (gradA[l][i] * gradD[j] + gradD[i] * gradA[l][j]);
    }
    
   }

   solveLinearSys(A, deltaT, b, nSys, w);
   fprintf(stderr,"DEBUG: singular values:\n");
   for(i=0; i<nSys; i++)
    fprintf(stderr," %e", w[i+1]);
   fprintf(stderr,"\n");
   
   normDbl(deltaT, (int) 8, &tmp);
   fprintf(stderr," Norm update: %e\n", tmp);
   for(k=0; k<8; k++)
    b[k] = deltaT[k];
/**
  rescale(b, (int) 8);
**/

   maxS = 1.0; minS = 1.0e-10;
   s = maxS;
   tmp = 0.0;
   while(s > minS) {
    for(k=0; k<8; k++)
     newT[k] = T[k] - s *b[k];
    residual(&newRes, newT, tPx, tPy, iPx, iPy, nPair);
    if (newRes < sumRes) {
     for(k=0; k<8; k++)
      T[k] = newT[k];
     sumRes = newRes;
     tmp = max(tmp, s);
    } else
     s /= 2.0;
   }
   
   fprintf(stderr," rational RMS = %e, step %e\n", sqrt(sumRes)/nPair, tmp);
   if (tmp == 0.0)
    break;
  }
  }
  
  /* Convert mapping to std form, with center postion tMean */
  moveRationalCenter((float)0.0, (float)0.0, T, (float)ixC, (float) iyC);

/* DEBUG: Check residual */
  fprintf(stderr,"Mapping:\n");
  fprintf(stderr," ixC, iyC = %f %f\n", ixC, iyC);
  fprintf(stderr," T[0], T[1] : %e %e\n", T[0], T[1]);
  fprintf(stderr," T[2] - T[5] : %e %e %e %e\n", T[2], T[3], T[4], T[5]);
  fprintf(stderr," T[6], T[7] : %e %e\n", T[6], T[7]);
  sumRes = 0.0;
  fprintf(stderr,"align:\n  tPx, tPy, iPx, iPy: mPx, mPy: res\n");
  for(k=0; k< nPair; k++) {
   rationalSample((float)tPx[k], (float)tPy[k], (float)ixC, (float)iyC, T, rPnt)
;
   res = (iPx[k] - rPnt[0]) *  (iPx[k] - rPnt[0]);
   res += (iPy[k] - rPnt[1]) *  (iPy[k] - rPnt[1]);
   fprintf(stderr,"%d %d %d %d : ", tPx[k], tPy[k], iPx[k], iPy[k]);
   fprintf(stderr,"%e %e : %e\n", rPnt[0], rPnt[1], sqrt(res));
   sumRes += res;
  }
  fprintf(stderr," RMS = %e\n", sqrt(sumRes)/nPair);

 }

residual(pRes, T, tPx, tPy, iPx, iPy, nPair)
 float *pRes, T[];
 int tPx[], tPy[], iPx[], iPy[];
 int nPair;
 {
  float sumRes, rPnt[2], res;
  int k;

  sumRes = 0.0;
  for(k=0; k< nPair; k++) {
    rationalSample((float)tPx[k], (float)tPy[k], (float)0.0, (float)0.0, T, rPnt);
    res = (iPx[k] - rPnt[0]) *  (iPx[k] - rPnt[0]);
    res += (iPy[k] - rPnt[1]) *  (iPy[k] - rPnt[1]);
    sumRes += res;
  }
  *pRes = sumRes;
 }
