#include <math.h>
#include "svd_D.h"

/*     this subroutine is a translation of the algol procedure svd, */
/*     num. math. 14, 403-420(1970) by golub and reinsch. */
/*     handbook for auto. comp., vol ii-linear algebra, 134-151(1971). */
/**    See http://www.netlib.org/  for Eispack svd.f Fortran77 version */

static double at,bt,ct;
#define PYTHAG(a,b) ((at=fabs(a)) > (bt=fabs(b)) ?\
(ct=bt/at,at*sqrt(1.0+ct*ct)) : (bt>0.0 ? (ct=at/bt,bt*sqrt(1.0+ct*ct)): 0.0))

static double maxarg1,maxarg2;
#define MAX(a,b) (maxarg1=(a),maxarg2=(b),(maxarg1) > (maxarg2) ?\
	(maxarg1) : (maxarg2))
#define SIGN(a,b) ((b) >= 0.0 ? fabs(a) : -fabs(a))

static int maxIts = 100;

int svdEispack_D(double a[][SVD_DIM], int m, int n,
               double w[SVD_DIM], double v[][SVD_DIM], double rv1[SVD_DIM])
{
	int flag,i,its,j,jj,k,l,nm;
	double c,f,h,s,x,y,z;
	double anorm=0.0,g=0.0,scale=0.0;
        double tst;
	void error();

	if (m < n) error("SVDCMP: You must augment A with extra zero rows","");
	for (i=1;i<=n;i++) {
		l=i+1;
		rv1[i]=scale*g;
		g=s=scale=0.0;
		if (i <= m) {
			for (k=i;k<=m;k++) scale += fabs(a[k][i]);
			if (scale) {
				for (k=i;k<=m;k++) {
					a[k][i] /= scale;
					s += a[k][i]*a[k][i];
				}
				f=a[i][i];
				g = -SIGN(sqrt(s),f);
				h=f*g-s;
				a[i][i]=f-g;
				if (i != n) {
					for (j=l;j<=n;j++) {
						for (s=0.0,k=i;k<=m;k++) s += a[k][i]*a[k][j];
						f=s/h;
						for (k=i;k<=m;k++) a[k][j] += f*a[k][i];
					}
				}
				for (k=i;k<=m;k++) a[k][i] *= scale;
			}
		}
		w[i]=scale*g;
		g=s=scale=0.0;
		if (i <= m && i != n) {
			for (k=l;k<=n;k++) scale += fabs(a[i][k]);
			if (scale) {
				for (k=l;k<=n;k++) {
					a[i][k] /= scale;
					s += a[i][k]*a[i][k];
				}
				f=a[i][l];
				g = -SIGN(sqrt(s),f);
				h=f*g-s;
				a[i][l]=f-g;
				for (k=l;k<=n;k++) rv1[k]=a[i][k]/h;
				if (i != m) {
					for (j=l;j<=m;j++) {
						for (s=0.0,k=l;k<=n;k++) s += a[j][k]*a[i][k];
						for (k=l;k<=n;k++) a[j][k] += s*rv1[k];
					}
				}
				for (k=l;k<=n;k++) a[i][k] *= scale;
			}
		}
		anorm=MAX(anorm,(fabs(w[i])+fabs(rv1[i])));
	}
	for (i=n;i>=1;i--) {
		if (i < n) {
			if (g) {
				for (j=l;j<=n;j++)
					v[j][i]=(a[i][j]/a[i][l])/g;
				for (j=l;j<=n;j++) {
					for (s=0.0,k=l;k<=n;k++) s += a[i][k]*v[k][j];
					for (k=l;k<=n;k++) v[k][j] += s*v[k][i];
				}
			}
			for (j=l;j<=n;j++) v[i][j]=v[j][i]=0.0;
		}
		v[i][i]=1.0;
		g=rv1[i];
		l=i;
	}
	for (i=n;i>=1;i--) {
		l=i+1;
		g=w[i];
		if (i < n)
			for (j=l;j<=n;j++) a[i][j]=0.0;
		if (g) {
			g=1.0/g;
			if (i != n) {
				for (j=l;j<=n;j++) {
					for (s=0.0,k=l;k<=m;k++) s += a[k][i]*a[k][j];
					f=(s/a[i][i])*g;
					for (k=i;k<=m;k++) a[k][j] += f*a[k][i];
				}
			}
			for (j=i;j<=m;j++) a[j][i] *= g;
		} else {
			for (j=i;j<=m;j++) a[j][i]=0.0;
		}
		++a[i][i];
	}

	for (k=n;k>=1;k--) {
		for (its=1;its<=maxIts;its++) {
			flag=1;
			for (l=k;l>=1;l--) {
				nm=l-1;
                                tst = fabs(rv1[l])+anorm;
				if (tst == anorm) {
					flag=0;
					break;
				}
                                tst = fabs(w[nm])+anorm;
				if (tst == anorm) break;
			}
                        /* Found a zero diagonal element w[nm] */
			if (flag) {
				c=0.0;
				s=1.0;
				for (i=l;i<=k;i++) {
					f=s*rv1[i];
                                        rv1[i] = c*rv1[i];
                                        tst = fabs(f)+anorm;
					if (tst == anorm) 
                                                break;
                                        else {
						g=w[i];
						h=PYTHAG(f,g);
						w[i]=h;
						h=1.0/h;
						c=g*h;
						s=(-f*h);
						for (j=1;j<=m;j++) {
							y=a[j][nm];
							z=a[j][i];
							a[j][nm]=y*c+z*s;
							a[j][i]=z*c-y*s;
						}
					}
				}
			}
			z=w[k];
			if (l == k) {
				if (z < 0.0) {
					w[k] = -z;
					for (j=1;j<=n;j++) v[j][k]=(-v[j][k]);
				}
				break;
			}
			if (its >= maxIts) {
                         return(-1);
                        }

			x=w[l];
			nm=k-1;
			y=w[nm];
			g=rv1[nm];
			h=rv1[k];
			f= 0.5*(((g+z)/h)*((g-z)/y)+y/h-h/y);
			g=PYTHAG(f,1.0);
                        f = x-(z/x)*z + (h/x) * (y/(f+SIGN(g,f))-h);
			c=s=1.0;
			for (j=l;j<=nm;j++) {
				i=j+1;
				g=rv1[i];
				y=w[i];
				h=s*g;
				g=c*g;
				z=PYTHAG(f,h);
				rv1[j]=z;
				c=f/z;
				s=h/z;
				f=x*c+g*s;
				g=g*c-x*s;
				h=y*s;
				y=y*c;
				for (jj=1;jj<=n;jj++) {
					x=v[jj][j];
					z=v[jj][i];
					v[jj][j]=x*c+z*s;
					v[jj][i]=z*c-x*s;
				}
				z=PYTHAG(f,h);
				w[j]=z;
				if (z != 0.0) {
					c=f/z;
					s=h/z;
				} 
				f=(c*g)+(s*y);
				x=(c*y)-(s*g);
				for (jj=1;jj<=m;jj++) {
					y=a[jj][j];
					z=a[jj][i];
					a[jj][j]=y*c+z*s;
					a[jj][i]=z*c-y*s;
				}
			}
			rv1[l]=0.0;
			rv1[k]=f;
			w[k]=x;
		}
	}
        return(0);
}

void sortSV_D(int *svPerm, double *w, int numSamp)
  /**********************************************
  Sort singular values into decreasing order,
  return permutation array svPerm *... ith sorted
  singular value is then w[svPerm[i]] 
  ***********************************************/
  {
   double tmp;
   int i, j, iTmp;
   
   for(i=1; i<=numSamp; i++)
    svPerm[i] = i;

   for(i=1; i<numSamp; i++) {
    /* find max in remaining set i..numSamp */
    tmp = w[svPerm[i]];
    iTmp = i;
    for(j=i+1; j<=numSamp; j++)
     if (w[svPerm[j]] > tmp) {
      tmp = w[svPerm[j]];
      iTmp = j;
     }
    /* Switch */
    j = svPerm[i];
    svPerm[i] = svPerm[iTmp];
    svPerm[iTmp] = j;
   }
    
  }

int solveLinearSys_D(double A[][SVD_DIM], double *x, double *b, 
                     int n, double *w)           
 {                                                             
   double V[SVD_DIM][SVD_DIM], scr[SVD_DIM], rv[SVD_DIM];
   int i,j,k;                                
                           
   k=svdEispack_D(A,n,n,w,V,rv);                                                         
   for(i=1; i<=n; i++) {      
    scr[i] = 0.0;
    for(j=1; j<=n; j++)          
     scr[i] += A[j][i] * b[j-1];                                   
    scr[i] /= w[i];                        
   }                                                    
                                   
   for(i=1; i<=n; i++) {        
    x[i-1] = 0.0;      
    for(j=1; j<=n; j++)                                            
     x[i-1] += V[i][j] * scr[j];
   }
   return(k);                  
 }
      
/* Given svd decomp, A = U w V^T, compute
   inverse of matrix A <- V w^-1 U^T */
void invertMatrix_D(double A[][SVD_DIM], int n, double w[SVD_DIM],
                  double U[][SVD_DIM], double V[][SVD_DIM])
 {                                                             
   double scr[SVD_DIM];
   int i,j,k;
                           
   for(k=1; k<=n; k++) {

    /* Compute scr = kth column of( w^-1 U^T) */
    for(i=1; i<=n; i++) {      
     scr[i] = U[k][i]/w[i];
    }                                                    
                                   
    /* Compute kth col of Ainv = V * scr */
    for(i=1; i<=n; i++) {        
     A[i][k] = 0.0;      
     for(j=1; j<=n; j++) 
      A[i][k] += V[i][j] * scr[j];
    }
   }

 }

#undef SIGN
#undef MAX
#undef PYTHAG
